import uuid

from dvc.testing.cloud import Cloud
from dvc.testing.path_info import CloudURLInfo

GS_BUCKET = "dvc-temp"


class GCP(Cloud, CloudURLInfo):

    IS_OBJECT_STORAGE = True

    def __init__(self, url, credentialpath):
        super().__init__(url)
        self.credentialpath = credentialpath

    def __truediv__(self, key):
        ret = super().__truediv__(key)
        ret.credentialpath = self.credentialpath
        return ret

    @property
    def config(self):
        # FIXME: we shouldn't need to set credentialpath if env var is set
        return {
            "url": self.url,
            "credentialpath": self.credentialpath,
        }

    @staticmethod
    def _get_storagepath():
        return GS_BUCKET + "/" + str(uuid.uuid4())

    @staticmethod
    def get_url():
        return "gs://" + GCP._get_storagepath()

    @property
    def _gc(self):
        from google.cloud.storage import Client

        return Client.from_service_account_json(self.credentialpath)

    @property
    def _bucket(self):
        return self._gc.bucket(self.bucket)

    @property
    def _blob(self):
        return self._bucket.blob(self.path)

    def is_file(self):
        if self.path.endswith("/"):
            return False

        return self._blob.exists()

    def is_dir(self):
        dir_path = self / ""
        return bool(list(self._bucket.list_blobs(prefix=dir_path.path)))

    def exists(self):
        return self.is_file() or self.is_dir()

    # pylint: disable-next=unused-argument
    def mkdir(self, mode=0o777, parents=False, exist_ok=False):
        assert mode == 0o777
        assert parents

    def write_bytes(self, contents):
        assert isinstance(contents, bytes)
        self._blob.upload_from_string(contents)

    def read_bytes(self):
        return self._blob.download_as_string()


class FakeGCP(GCP):
    """Fake GCP client that is supposed to be using a mock server's endpoint"""

    def __init__(self, url, endpoint_url: str):
        super().__init__(
            url, ""  # no need to provide credentials for a mock server
        )
        self.endpoint_url = endpoint_url

    def __truediv__(self, key):
        ret = super().__truediv__(key)
        ret.endpoint_url = self.endpoint_url
        return ret

    @property
    def config(self):
        return {"url": self.url, "endpointurl": self.endpoint_url}

    def get_url(self):  # pylint: disable=arguments-differ
        return self.url

    @property
    def _gc(self):
        from google.cloud.storage import Client

        return Client(
            use_auth_w_custom_endpoint=False,
            client_options={"api_endpoint": self.endpoint_url},
        )



from typing import Optional, Tuple

from dvc_objects.fs.base import AnyFSPath
from dvc_objects.fs.path import Path


class GSPath(Path):
    def split_version(self, path: AnyFSPath) -> Tuple[str, Optional[str]]:
        from gcsfs import GCSFileSystem

        parts = GCSFileSystem._split_path(  # pylint: disable=protected-access
            path, version_aware=True
        )
        bucket, key, generation = parts
        return f"gs://{bucket}/{key}", generation

    def join_version(self, path: AnyFSPath, version_id: Optional[str]) -> str:
        path, path_version = self.split_version(path)
        if path_version:
            raise ValueError("path already includes an object generation")
        return f"{path}#{version_id}" if version_id else path

    def version_path(self, path: AnyFSPath, version_id: Optional[str]) -> str:
        path, _ = self.split_version(path)
        return self.join_version(path, version_id)

    def coalesce_version(
        self, path: AnyFSPath, version_id: Optional[str]
    ) -> Tuple[AnyFSPath, Optional[str]]:
        path, path_version_id = self.split_version(path)
        versions = {ver for ver in (version_id, path_version_id) if ver}
        if len(versions) > 1:
            raise ValueError(
                f"Path version mismatch: '{path}', '{version_id}'"
            )
        return path, (versions.pop() if versions else None)



import threading

# pylint:disable=abstract-method
from dvc_objects.fs.base import ObjectFileSystem
from funcy import cached_property, wrap_prop

from .path import GSPath


class GSFileSystem(ObjectFileSystem):
    protocol = "gs"
    REQUIRES = {"gcsfs": "gcsfs"}
    PARAM_CHECKSUM = "etag"

    @cached_property
    def path(self) -> GSPath:
        def _getcwd():
            return self.fs.root_marker

        return GSPath(self.sep, getcwd=_getcwd)

    def _prepare_credentials(self, **config):
        login_info = {"consistency": None}
        login_info["version_aware"] = config.get("version_aware", False)
        login_info["project"] = config.get("projectname")
        login_info["token"] = config.get("credentialpath")
        login_info["endpoint_url"] = config.get("endpointurl")
        login_info["session_kwargs"] = {"trust_env": True}
        return login_info

    @wrap_prop(threading.Lock())
    @cached_property
    def fs(self):
        from gcsfs import GCSFileSystem

        return GCSFileSystem(**self.fs_args)

    @classmethod
    def _strip_protocol(cls, path: str) -> str:
        from fsspec.utils import infer_storage_options

        return infer_storage_options(path)["path"]

    def unstrip_protocol(self, path: str) -> str:
        return "gs://" + path.lstrip("/")

    @staticmethod
    def _get_kwargs_from_urls(urlpath: str):
        from gcsfs import GCSFileSystem

        parts = GCSFileSystem._split_path(  # pylint: disable=protected-access
            urlpath, version_aware=True
        )
        _, _, generation = parts
        if generation is not None:
            return {"version_aware": True}
        return {}



import pytest
from dvc.testing.api_tests import (  # noqa, pylint: disable=unused-import
    TestAPI,
)
from dvc.testing.remote_tests import (  # noqa, pylint: disable=unused-import
    TestRemote,
)
from dvc.testing.workspace_tests import (  # noqa, pylint: disable=unused-import
    TestGetUrl,
)
from dvc.testing.workspace_tests import TestImport as _TestImport
from dvc.testing.workspace_tests import (  # noqa, pylint: disable=unused-import
    TestLsUrl,
)


@pytest.fixture
def cloud(make_cloud):
    yield make_cloud(typ="gs")


@pytest.fixture
def remote(make_remote):
    yield make_remote(name="upstream", typ="gs")


@pytest.fixture
def workspace(make_workspace):
    yield make_workspace(name="workspace", typ="gs")


class TestImport(_TestImport):
    @pytest.fixture
    def stage_md5(self):
        # stage md5 depends on the dir_md5, see below
        return None

    @pytest.fixture
    def is_object_storage(self):
        return True

    @pytest.fixture
    def dir_md5(self):
        # dir md5 for imported dirs changes on every run.
        # we temporarily set it to None until the test is revisited
        # https://github.com/iterative/dvc-gs/issues/7#issuecomment-1218497067
        return None



import os

import pytest

from .cloud import GCP, FakeGCP


@pytest.fixture(scope="session")
def gs_creds(tmp_path_factory):
    gcreds_path = os.getenv("GOOGLE_APPLICATION_CREDENTIALS")
    if not gcreds_path:
        dpath = tmp_path_factory.mktemp("gs_creds")
        fpath = dpath / "creds.json"
        fpath.write_text(os.getenv("GS_CREDS"))
        gcreds_path = os.fspath(fpath)
    # FIXME: we shouldn't need to use this return value
    # if GOOGLE_APPLICATION_CREDENTIALS is set
    return gcreds_path


@pytest.fixture
def make_gs(tmp_gcs_path, fake_gcs_server):
    def _make_gs():
        return FakeGCP(
            str(tmp_gcs_path).replace("gcs://", "gs://"),
            endpoint_url=fake_gcs_server,
        )

    return _make_gs


@pytest.fixture
def gs(make_gs):  # pylint: disable=redefined-outer-name
    return make_gs()


@pytest.fixture
def real_gs():
    yield GCP(GCP.get_url(), "")
